'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var combineArrays = _interopDefault(require('combine-arrays'));
var pathToRegexpWithReversibleKeys = _interopDefault(require('path-to-regexp-with-reversible-keys'));
var thenDenodeify = _interopDefault(require('then-denodeify'));
var eventemitter3 = _interopDefault(require('eventemitter3'));
var hashBrownRouter = _interopDefault(require('hash-brown-router'));
var pagePathBuilder = _interopDefault(require('page-path-builder'));
var isoNextTick = _interopDefault(require('iso-next-tick'));

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var stateStringParser = createCommonjsModule(function (module) {
	module.exports = function (stateString) {
		return stateString.split('.').reduce(function (stateNames, latestNameChunk) {
			stateNames.push(stateNames.length ? stateNames[stateNames.length - 1] + '.' + latestNameChunk : latestNameChunk);

			return stateNames;
		}, []);
	};
});

var stateState = function StateState() {
	var states = {};

	function getHierarchy(name) {
		var names = stateStringParser(name);

		return names.map(function (name) {
			if (!states[name]) {
				throw new Error('State ' + name + ' not found');
			}
			return states[name];
		});
	}

	function getParent(name) {
		var parentName = getParentName(name);

		return parentName && states[parentName];
	}

	function getParentName(name) {
		var names = stateStringParser(name);

		if (names.length > 1) {
			var secondToLast = names.length - 2;

			return names[secondToLast];
		} else {
			return null;
		}
	}

	function guaranteeAllStatesExist(newStateName) {
		var stateNames = stateStringParser(newStateName);
		var statesThatDontExist = stateNames.filter(function (name) {
			return !states[name];
		});

		if (statesThatDontExist.length > 0) {
			throw new Error('State ' + statesThatDontExist[statesThatDontExist.length - 1] + ' does not exist');
		}
	}

	function buildFullStateRoute(stateName) {
		return getHierarchy(stateName).map(function (state) {
			return '/' + (state.route || '');
		}).join('').replace(/\/{2,}/g, '/');
	}

	function applyDefaultChildStates(stateName) {
		var state = states[stateName];

		var defaultChildStateName = state && (typeof state.defaultChild === 'function' ? state.defaultChild() : state.defaultChild);

		if (!defaultChildStateName) {
			return stateName;
		}

		var fullStateName = stateName + '.' + defaultChildStateName;

		return applyDefaultChildStates(fullStateName);
	}

	return {
		add: function add(name, state) {
			states[name] = state;
		},
		get: function get(name) {
			return name && states[name];
		},

		getHierarchy: getHierarchy,
		getParent: getParent,
		getParentName: getParentName,
		guaranteeAllStatesExist: guaranteeAllStatesExist,
		buildFullStateRoute: buildFullStateRoute,
		applyDefaultChildStates: applyDefaultChildStates
	};
};

var extend = function extend() {
  for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  return Object.assign.apply(Object, [{}].concat(args));
};

var stateComparison_1 = function StateComparison(stateState) {
	var getPathParameters = pathParameters();

	var parametersChanged = function parametersChanged(args) {
		return parametersThatMatterWereChanged(extend(args, { stateState: stateState, getPathParameters: getPathParameters }));
	};

	return function (args) {
		return stateComparison(extend(args, { parametersChanged: parametersChanged }));
	};
};

function pathParameters() {
	var parameters = {};

	return function (path) {
		if (!path) {
			return [];
		}

		if (!parameters[path]) {
			parameters[path] = pathToRegexpWithReversibleKeys(path).keys.map(function (key) {
				return key.name;
			});
		}

		return parameters[path];
	};
}

function parametersThatMatterWereChanged(_ref) {
	var stateState = _ref.stateState,
	    getPathParameters = _ref.getPathParameters,
	    stateName = _ref.stateName,
	    fromParameters = _ref.fromParameters,
	    toParameters = _ref.toParameters;

	var state = stateState.get(stateName);
	var querystringParameters = state.querystringParameters || [];
	var parameters = getPathParameters(state.route).concat(querystringParameters);

	return Array.isArray(parameters) && parameters.some(function (key) {
		return fromParameters[key] !== toParameters[key];
	});
}

function stateComparison(_ref2) {
	var parametersChanged = _ref2.parametersChanged,
	    original = _ref2.original,
	    destination = _ref2.destination;

	var states = combineArrays({
		start: stateStringParser(original.name),
		end: stateStringParser(destination.name)
	});

	return states.map(function (_ref3) {
		var start = _ref3.start,
		    end = _ref3.end;
		return {
			nameBefore: start,
			nameAfter: end,
			stateNameChanged: start !== end,
			stateParametersChanged: start === end && parametersChanged({
				stateName: start,
				fromParameters: original.parameters,
				toParameters: destination.parameters
			})
		};
	});
}

var currentState = function CurrentState() {
	var current = {
		name: '',
		parameters: {}
	};

	return {
		get: function get() {
			return current;
		},
		set: function set(name, parameters) {
			current = {
				name: name,
				parameters: parameters
			};
		}
	};
};

var stateChangeLogic = function stateChangeLogic(stateComparisonResults) {
	var hitChangingState = false;
	var hitDestroyedState = false;

	var output = {
		destroy: [],
		change: [],
		create: []
	};

	stateComparisonResults.forEach(function (state) {
		hitChangingState = hitChangingState || state.stateParametersChanged;
		hitDestroyedState = hitDestroyedState || state.stateNameChanged;

		if (state.nameBefore) {
			if (hitDestroyedState) {
				output.destroy.push(state.nameBefore);
			} else if (hitChangingState) {
				output.change.push(state.nameBefore);
			}
		}

		if (state.nameAfter && hitDestroyedState) {
			output.create.push(state.nameAfter);
		}
	});

	return output;
};

var stateTransitionManager = function stateTransitionManager(emitter) {
	var currentTransitionAttempt = null;
	var nextTransition = null;

	function doneTransitioning() {
		currentTransitionAttempt = null;
		if (nextTransition) {
			beginNextTransitionAttempt();
		}
	}

	var isTransitioning = function isTransitioning() {
		return !!currentTransitionAttempt;
	};

	function beginNextTransitionAttempt() {
		currentTransitionAttempt = nextTransition;
		nextTransition = null;
		currentTransitionAttempt.beginStateChange();
	}

	function cancelCurrentTransition() {
		currentTransitionAttempt.transition.cancelled = true;
		var err = new Error('State transition cancelled by the state transition manager');
		err.wasCancelledBySomeoneElse = true;
		emitter.emit('stateChangeCancelled', err);
	}

	emitter.on('stateChangeAttempt', function (beginStateChange) {
		nextTransition = createStateTransitionAttempt(beginStateChange);

		if (isTransitioning() && currentTransitionAttempt.transition.cancellable) {
			cancelCurrentTransition();
		} else if (!isTransitioning()) {
			beginNextTransitionAttempt();
		}
	});

	emitter.on('stateChangeError', doneTransitioning);
	emitter.on('stateChangeCancelled', doneTransitioning);
	emitter.on('stateChangeEnd', doneTransitioning);

	function createStateTransitionAttempt(_beginStateChange) {
		var transition = {
			cancelled: false,
			cancellable: true
		};
		return {
			transition: transition,
			beginStateChange: function beginStateChange() {
				for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
					args[_key] = arguments[_key];
				}

				return _beginStateChange.apply(undefined, [transition].concat(args));
			}
		};
	}
};

var defaultRouterOptions = { reverse: false };

// Pulled from https://github.com/joliss/promise-map-series and prettied up a bit

var promiseMapSeries = function sequence(array, iterator) {
	var currentPromise = Promise.resolve();
	return Promise.all(array.map(function (value, i) {
		return currentPromise = currentPromise.then(function () {
			return iterator(value, i, array);
		});
	}));
};

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};

var getProperty = function getProperty(name) {
	return function (obj) {
		return obj[name];
	};
};
var reverse = function reverse(ary) {
	return ary.slice().reverse();
};
var isFunction = function isFunction(property) {
	return function (obj) {
		return typeof obj[property] === 'function';
	};
};
var isThenable = function isThenable(object) {
	return object && ((typeof object === 'undefined' ? 'undefined' : _typeof(object)) === 'object' || typeof object === 'function') && typeof object.then === 'function';
};
var promiseMe = function promiseMe(fn) {
	for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
		args[_key - 1] = arguments[_key];
	}

	return new Promise(function (resolve) {
		return resolve(fn.apply(undefined, args));
	});
};

var expectedPropertiesOfAddState = ['name', 'route', 'defaultChild', 'data', 'template', 'resolve', 'activate', 'querystringParameters', 'defaultQuerystringParameters', 'defaultParameters'];

var abstractStateRouter = function StateProvider(makeRenderer, rootElement, stateRouterOptions) {
	var prototypalStateHolder = stateState();
	var lastCompletelyLoadedState = currentState();
	var lastStateStartedActivating = currentState();
	var stateProviderEmitter = new eventemitter3();
	var compareStartAndEndStates = stateComparison_1(prototypalStateHolder);

	var stateNameToArrayofStates = function stateNameToArrayofStates(stateName) {
		return stateStringParser(stateName).map(prototypalStateHolder.get);
	};

	stateTransitionManager(stateProviderEmitter);

	var _extend = extend({
		throwOnError: true,
		pathPrefix: '#'
	}, stateRouterOptions),
	    throwOnError = _extend.throwOnError,
	    pathPrefix = _extend.pathPrefix;

	var router = stateRouterOptions.router || hashBrownRouter(defaultRouterOptions);

	router.on('not found', function (route, parameters) {
		stateProviderEmitter.emit('routeNotFound', route, parameters);
	});

	var destroyDom = null;
	var getDomChild = null;
	var renderDom = null;
	var resetDom = null;

	var activeStateResolveContent = {};
	var activeDomApis = {};
	var activeEmitters = {};

	function handleError(event, err) {
		isoNextTick(function () {
			stateProviderEmitter.emit(event, err);
			console.error(event + ' - ' + err.message);
			if (throwOnError) {
				throw err;
			}
		});
	}

	function destroyStateName(stateName) {
		var state = prototypalStateHolder.get(stateName);
		stateProviderEmitter.emit('beforeDestroyState', {
			state: state,
			domApi: activeDomApis[stateName]
		});

		activeEmitters[stateName].emit('destroy');
		activeEmitters[stateName].removeAllListeners();
		delete activeEmitters[stateName];
		delete activeStateResolveContent[stateName];

		return destroyDom(activeDomApis[stateName]).then(function () {
			delete activeDomApis[stateName];
			stateProviderEmitter.emit('afterDestroyState', {
				state: state
			});
		});
	}

	function resetStateName(parameters, stateName) {
		var domApi = activeDomApis[stateName];
		var content = getContentObject(activeStateResolveContent, stateName);
		var state = prototypalStateHolder.get(stateName);

		stateProviderEmitter.emit('beforeResetState', {
			domApi: domApi,
			content: content,
			state: state,
			parameters: parameters
		});

		activeEmitters[stateName].emit('destroy');
		delete activeEmitters[stateName];

		return resetDom({
			domApi: domApi,
			content: content,
			template: state.template,
			parameters: parameters
		}).then(function (newDomApi) {
			if (newDomApi) {
				activeDomApis[stateName] = newDomApi;
			}

			stateProviderEmitter.emit('afterResetState', {
				domApi: activeDomApis[stateName],
				content: content,
				state: state,
				parameters: parameters
			});
		});
	}

	function getChildElementForStateName(stateName) {
		return new Promise(function (resolve) {
			var parent = prototypalStateHolder.getParent(stateName);
			if (parent) {
				var parentDomApi = activeDomApis[parent.name];
				resolve(getDomChild(parentDomApi));
			} else {
				resolve(rootElement);
			}
		});
	}

	function renderStateName(parameters, stateName) {
		return getChildElementForStateName(stateName).then(function (element) {
			var state = prototypalStateHolder.get(stateName);
			var content = getContentObject(activeStateResolveContent, stateName);

			stateProviderEmitter.emit('beforeCreateState', {
				state: state,
				content: content,
				parameters: parameters
			});

			return renderDom({
				template: state.template,
				element: element,
				content: content,
				parameters: parameters
			}).then(function (domApi) {
				activeDomApis[stateName] = domApi;
				stateProviderEmitter.emit('afterCreateState', {
					state: state,
					domApi: domApi,
					content: content,
					parameters: parameters
				});
				return domApi;
			});
		});
	}

	function renderAll(stateNames, parameters) {
		return promiseMapSeries(stateNames, function (stateName) {
			return renderStateName(parameters, stateName);
		});
	}

	function onRouteChange(state, parameters) {
		try {
			var finalDestinationStateName = prototypalStateHolder.applyDefaultChildStates(state.name);

			if (finalDestinationStateName === state.name) {
				emitEventAndAttemptStateChange(finalDestinationStateName, parameters);
			} else {
				// There are default child states that need to be applied

				var theRouteWeNeedToEndUpAt = makePath(finalDestinationStateName, parameters);
				var currentRoute = router.location.get();

				if (theRouteWeNeedToEndUpAt === currentRoute) {
					// the child state has the same route as the current one, just start navigating there
					emitEventAndAttemptStateChange(finalDestinationStateName, parameters);
				} else {
					// change the url to match the full default child state route
					stateProviderEmitter.go(finalDestinationStateName, parameters, { replace: true });
				}
			}
		} catch (err) {
			handleError('stateError', err);
		}
	}

	function addState(state) {
		if (typeof state === 'undefined') {
			throw new Error('Expected \'state\' to be passed in.');
		} else if (typeof state.name === 'undefined') {
			throw new Error('Expected the \'name\' option to be passed in.');
		} else if (typeof state.template === 'undefined') {
			throw new Error('Expected the \'template\' option to be passed in.');
		}
		Object.keys(state).filter(function (key) {
			return expectedPropertiesOfAddState.indexOf(key) === -1;
		}).forEach(function (key) {
			console.warn('Unexpected property passed to addState:', key);
		});

		prototypalStateHolder.add(state.name, state);

		var route = prototypalStateHolder.buildFullStateRoute(state.name);

		router.add(route, function (parameters) {
			return onRouteChange(state, parameters);
		});
	}

	function getStatesToResolve(stateChanges) {
		return stateChanges.change.concat(stateChanges.create).map(prototypalStateHolder.get);
	}

	function emitEventAndAttemptStateChange(newStateName, parameters) {
		stateProviderEmitter.emit('stateChangeAttempt', function stateGo(transition) {
			attemptStateChange(newStateName, parameters, transition);
		});
	}

	function attemptStateChange(newStateName, parameters, transition) {
		function ifNotCancelled(fn) {
			return function () {
				if (transition.cancelled) {
					var err = new Error('The transition to ' + newStateName + ' was cancelled');
					err.wasCancelledBySomeoneElse = true;
					throw err;
				} else {
					return fn.apply(undefined, arguments);
				}
			};
		}

		return promiseMe(prototypalStateHolder.guaranteeAllStatesExist, newStateName).then(function applyDefaultParameters() {
			var state = prototypalStateHolder.get(newStateName);
			var defaultParams = state.defaultParameters || state.defaultQuerystringParameters || {};
			var needToApplyDefaults = Object.keys(defaultParams).some(function missingParameterValue(param) {
				return typeof parameters[param] === 'undefined';
			});

			if (needToApplyDefaults) {
				throw redirector(newStateName, extend(defaultParams, parameters));
			}
			return state;
		}).then(ifNotCancelled(function (state) {
			stateProviderEmitter.emit('stateChangeStart', state, parameters, stateNameToArrayofStates(state.name));
			lastStateStartedActivating.set(state.name, parameters);
		})).then(function getStateChanges() {
			var stateComparisonResults = compareStartAndEndStates({
				original: lastCompletelyLoadedState.get(),
				destination: {
					name: newStateName,
					parameters: parameters
				}
			});
			return stateChangeLogic(stateComparisonResults); // { destroy, change, create }
		}).then(ifNotCancelled(function resolveDestroyAndActivateStates(stateChanges) {
			return resolveStates(getStatesToResolve(stateChanges), extend(parameters)).catch(function onResolveError(e) {
				e.stateChangeError = true;
				throw e;
			}).then(ifNotCancelled(function destroyAndActivate(stateResolveResultsObject) {
				transition.cancellable = false;

				var activateAll = function activateAll() {
					return activateStates(stateChanges.change.concat(stateChanges.create));
				};

				activeStateResolveContent = extend(activeStateResolveContent, stateResolveResultsObject);

				return promiseMapSeries(reverse(stateChanges.destroy), destroyStateName).then(function () {
					return promiseMapSeries(reverse(stateChanges.change), function (stateName) {
						return resetStateName(extend(parameters), stateName);
					});
				}).then(function () {
					return renderAll(stateChanges.create, extend(parameters)).then(activateAll);
				});
			}));

			function activateStates(stateNames) {
				return stateNames.map(prototypalStateHolder.get).forEach(function (state) {
					var emitter = new eventemitter3();
					var context = Object.create(emitter);
					context.domApi = activeDomApis[state.name];
					context.data = state.data;
					context.parameters = parameters;
					context.content = getContentObject(activeStateResolveContent, state.name);
					activeEmitters[state.name] = emitter;

					try {
						state.activate && state.activate(context);
					} catch (e) {
						isoNextTick(function () {
							throw e;
						});
					}
				});
			}
		})).then(function stateChangeComplete() {
			lastCompletelyLoadedState.set(newStateName, parameters);
			try {
				stateProviderEmitter.emit('stateChangeEnd', prototypalStateHolder.get(newStateName), parameters, stateNameToArrayofStates(newStateName));
			} catch (e) {
				handleError('stateError', e);
			}
		}).catch(ifNotCancelled(function handleStateChangeError(err) {
			if (err && err.redirectTo) {
				stateProviderEmitter.emit('stateChangeCancelled', err);
				return stateProviderEmitter.go(err.redirectTo.name, err.redirectTo.params, { replace: true });
			} else if (err) {
				handleError('stateChangeError', err);
			}
		})).catch(function handleCancellation(err) {
			if (err && err.wasCancelledBySomeoneElse) {
				// we don't care, the state transition manager has already emitted the stateChangeCancelled for us
			} else {
				throw new Error('This probably shouldn\'t happen, maybe file an issue or something ' + err);
			}
		});
	}

	function makePath(stateName, parameters, options) {
		function getGuaranteedPreviousState() {
			if (!lastStateStartedActivating.get().name) {
				throw new Error('makePath required a previous state to exist, and none was found');
			}
			return lastStateStartedActivating.get();
		}
		if (options && options.inherit) {
			parameters = extend(getGuaranteedPreviousState().parameters, parameters);
		}

		var destinationStateName = stateName === null ? getGuaranteedPreviousState().name : stateName;

		var destinationState = prototypalStateHolder.get(destinationStateName) || {};
		var defaultParams = destinationState.defaultParameters || destinationState.defaultQuerystringParameters;

		parameters = extend(defaultParams, parameters);

		prototypalStateHolder.guaranteeAllStatesExist(destinationStateName);
		var route = prototypalStateHolder.buildFullStateRoute(destinationStateName);
		return pagePathBuilder(route, parameters || {});
	}

	var defaultOptions = {
		replace: false
	};

	stateProviderEmitter.addState = addState;
	stateProviderEmitter.go = function (newStateName, parameters, options) {
		options = extend(defaultOptions, options);
		var goFunction = options.replace ? router.replace : router.go;

		return promiseMe(makePath, newStateName, parameters, options).then(goFunction, function (err) {
			return handleError('stateChangeError', err);
		});
	};
	stateProviderEmitter.evaluateCurrentRoute = function (defaultState, defaultParams) {
		return promiseMe(makePath, defaultState, defaultParams).then(function (defaultPath) {
			router.evaluateCurrent(defaultPath);
		}).catch(function (err) {
			return handleError('stateError', err);
		});
	};
	stateProviderEmitter.makePath = function (stateName, parameters, options) {
		return pathPrefix + makePath(stateName, parameters, options);
	};
	stateProviderEmitter.stateIsActive = function (stateName) {
		var parameters = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

		var currentState$$1 = lastCompletelyLoadedState.get();
		var stateNameMatches = currentState$$1.name === stateName || currentState$$1.name.indexOf(stateName + '.') === 0;
		var parametersWereNotPassedIn = !parameters;

		return stateNameMatches && (parametersWereNotPassedIn || Object.keys(parameters).every(function (key) {
			return parameters[key] === currentState$$1.parameters[key];
		}));
	};

	var renderer = makeRenderer(stateProviderEmitter);

	destroyDom = thenDenodeify(renderer.destroy);
	getDomChild = thenDenodeify(renderer.getChildElement);
	renderDom = thenDenodeify(renderer.render);
	resetDom = thenDenodeify(renderer.reset);

	return stateProviderEmitter;
};

function getContentObject(stateResolveResultsObject, stateName) {
	var allPossibleResolvedStateNames = stateStringParser(stateName);

	return allPossibleResolvedStateNames.filter(function (stateName) {
		return stateResolveResultsObject[stateName];
	}).reduce(function (obj, stateName) {
		return extend(obj, stateResolveResultsObject[stateName]);
	}, {});
}

function redirector(newStateName, parameters) {
	return {
		redirectTo: {
			name: newStateName,
			params: parameters
		}
	};
}

// { [stateName]: resolveResult }
function resolveStates(states, parameters) {
	var statesWithResolveFunctions = states.filter(isFunction('resolve'));
	var stateNamesWithResolveFunctions = statesWithResolveFunctions.map(getProperty('name'));

	var resolves = Promise.all(statesWithResolveFunctions.map(function (state) {
		return new Promise(function (resolve, reject) {
			var resolveCb = function resolveCb(err, content) {
				return err ? reject(err) : resolve(content);
			};

			resolveCb.redirect = function (newStateName, parameters) {
				reject(redirector(newStateName, parameters));
			};

			var res = state.resolve(state.data, parameters, resolveCb);
			if (isThenable(res)) {
				resolve(res);
			}
		});
	}));

	return resolves.then(function (resolveResults) {
		return combineArrays({
			stateName: stateNamesWithResolveFunctions,
			resolveResult: resolveResults
		}).reduce(function (obj, result) {
			obj[result.stateName] = result.resolveResult;
			return obj;
		}, {});
	});
}

module.exports = abstractStateRouter;
//# sourceMappingURL=bundle.js.map
